"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, os, re, unittest, tempfile, shutil, pathlib

path = pathlib.Path(__file__).parent / '..' / 'data'

fname1 = str(path / 'nucl_data.fas')
fname2 = str(path / 'prot_data.fas')
fname3 = str(path / 'simul-cds.fas')
mask1 = str(path / 'mask_nucl_dustmasker.asnb')
mask2 = str(path / 'mask_nucl_windowmasker.asnb')

class makeblastdb_test(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.dnadb = egglib.io.from_fasta(fname1, alphabet=egglib.alphabets.DNA)
        self.protdb = egglib.io.from_fasta(fname2, alphabet=egglib.alphabets.protein)

    def tmpf(self, fname):
        return pathlib.Path(self.tmp) / fname

    def tearDown(self):
        shutil.rmtree(self.tmp)

    def test_dbtype(self):
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.protdb, dbtype='nucl')
        self.assertIn('alphabet/dbtype mismatch', str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.dnadb, dbtype='prot')
        self.assertIn('alphabet/dbtype mismatch', str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.dnadb, dbtype='prout')
        self.assertIn('alphabet/dbtype mismatch', str(cm.exception))

    def test_out(self):
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.dnadb)
        self.assertIn('`out` is required', str(cm.exception))
        with self.assertRaises(TypeError):
            egglib.wrappers.makeblastdb(self.dnadb, out=100)

        # successfull database creation
        egglib.wrappers.makeblastdb(self.dnadb, out=self.tmpf('test1'))
        self.assertTrue(self.tmpf('test1.nhr').is_file())
        self.assertTrue(self.tmpf('test1.nsq').is_file())
        self.assertTrue(self.tmpf('test1.nin').is_file())

        egglib.wrappers.makeblastdb(self.protdb, out=self.tmpf('test2'))
        self.assertTrue(self.tmpf('test2.phr').is_file())
        self.assertTrue(self.tmpf('test2.psq').is_file())
        self.assertTrue(self.tmpf('test2.pin').is_file())

        shutil.copy(fname1, self.tmpf('a.fas'))
        egglib.wrappers.makeblastdb(self.tmpf('a.fas'), dbtype='nucl')
        self.assertTrue(self.tmpf('a.fas.nhr').is_file())
        self.assertTrue(self.tmpf('a.fas.nsq').is_file())
        self.assertTrue(self.tmpf('a.fas.nin').is_file())

        # error in case blastdb is used: must be different name
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.tmpf('a.fas'), dbtype='nucl', input_type='blastdb')
        self.assertIn('a different database name is required for input type `blastdb`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.tmpf('a.fas'), dbtype='nucl', input_type='blastdb', out=self.tmpf('a.fas'))
        self.assertIn('a different database name is required for input type `blastdb`', str(cm.exception))

        egglib.wrappers.makeblastdb(self.tmpf('a.fas'), dbtype='nucl', input_type='blastdb', out=self.tmpf('b.fas'))
        self.assertTrue(self.tmpf('b.fas.nhr').is_file())
        self.assertTrue(self.tmpf('b.fas.nsq').is_file())
        self.assertTrue(self.tmpf('b.fas.nin').is_file())

    def test_source(self):
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb('non-existent')
        self.assertIn('file not found', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1)
        self.assertIn('`dbtype` is required', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'))
        self.assertTrue(self.tmpf('test1.nhr').is_file())
        self.assertTrue(self.tmpf('test1.nsq').is_file())
        self.assertTrue(self.tmpf('test1.nin').is_file())

        egglib.wrappers.makeblastdb(fname2, dbtype='prot', out=self.tmpf('test2'))
        self.assertTrue(self.tmpf('test2.phr').is_file())
        self.assertTrue(self.tmpf('test2.psq').is_file())
        self.assertTrue(self.tmpf('test2.pin').is_file())

    def test_input_type(self):
        egglib.wrappers.makeblastdb(self.dnadb, out=self.tmpf('test1'))
        self.assertTrue(self.tmpf('test1.nhr').is_file())
        self.assertTrue(self.tmpf('test1.nsq').is_file())
        self.assertTrue(self.tmpf('test1.nin').is_file())

        egglib.wrappers.makeblastdb(self.dnadb, out=self.tmpf('test2'), input_type='fasta')
        self.assertTrue(self.tmpf('test2.nhr').is_file())
        self.assertTrue(self.tmpf('test2.nsq').is_file())
        self.assertTrue(self.tmpf('test2.nin').is_file())

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(self.dnadb, out=self.tmpf('test1'), input_type='asn1_bin')
        self.assertIn('`input_type` must be "fasta"', str(cm.exception))

        with self.assertRaises(RuntimeError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', input_type='asn1_bin')
        self.assertIn('does not match input format type', str(cm.exception))

        with self.assertRaises(RuntimeError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', input_type='asn1_txt')
        self.assertIn('does not match input format type', str(cm.exception))

        with self.assertRaises(RuntimeError) as cm:
            egglib.wrappers.makeblastdb(self.tmpf('test_non_existent'), dbtype='nucl', input_type='blastdb', out=self.tmpf('prout'))
        self.assertIn('No alias or index file found for nucleotide database', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', input_type='confuse_a_cat')
        self.assertIn('invalid value for `input_type`', str(cm.exception))

    def test_title(self):
        stdout = egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'))
        self.assertIsNotNone(re.search('New DB title: +' + re.escape(fname1), stdout))

        stdout = egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), title='A Database Title')
        self.assertIsNotNone(re.search('New DB title: +A Database Title', stdout))

        stdout = egglib.wrappers.makeblastdb(self.protdb, out=self.tmpf('test1'))
        self.assertIsNotNone(re.search('New DB title: +prot database from an EggLib Container', stdout))

    def test_parse_seqids(self):
        for i in self.dnadb: i.name = 'seq'
        with self.assertRaises(RuntimeError) as cm:
            egglib.wrappers.makeblastdb(self.dnadb, dbtype='nucl', out=self.tmpf('test1'), parse_seqids=True)
        self.assertIn('Duplicate seq_ids are found', str(cm.exception))
        egglib.wrappers.makeblastdb(self.dnadb, dbtype='nucl', out=self.tmpf('test1'), parse_seqids=False)

    def test_hash_index(self):
        egglib.wrappers.makeblastdb(self.dnadb, dbtype='nucl', out=self.tmpf('test2'), hash_index=True)
        self.assertTrue(self.tmpf('test2.nhi').is_file())

    def test_mask_data(self):
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'))
        with self.assertRaises(TypeError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=mask1)
        self.assertIn('`mask_data` must be a list or a tuple', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[])
        self.assertIn('there must be at least one item in `mask_data`', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1], parse_seqids=True)
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], parse_seqids=True)
            # it is required to use parse_seqids because it has been used for creating masks

    def test_mask_id(self):
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_id=[])
        self.assertIn('`mask_id` requires `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_id=['dustmasker', 'windowmasker'])
        self.assertIn('`mask_id` requires `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=[])
        self.assertIn('`mask_id` must have the same length than `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker'])
        self.assertIn('`mask_id` must have the same length than `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker', 'napolmasker'])
        self.assertIn('`mask_id` must have the same length than `mask_data`', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1], mask_id=['dustmasker'], parse_seqids=True)
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker'], parse_seqids=True)

    def test_mask_desc(self):
        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_desc=[])
        self.assertIn('`mask_desc` requires `mask_id`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_id=['a'], mask_desc=['a'])
        self.assertIn('`mask_id` requires `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker'], mask_desc=[])
        self.assertIn('`mask_desc` must have the same length than `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker'], mask_desc=['dustmasker'])
        self.assertIn('`mask_desc` must have the same length than `mask_data`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker'], mask_desc=['dustmasker', 'windowmasker', 'napolmasker'], parse_seqids=True)
        self.assertIn('`mask_desc` must have the same length than `mask_data`', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1], mask_id=['dustmasker'], mask_desc=['a mask based on dust'], parse_seqids=True)
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), mask_data=[mask1, mask2], mask_id=['dustmasker', 'windowmasker'], mask_desc=['a mask based on dust', 'a mask based on windows'], parse_seqids=True)

    def test_blastdb_version(self):
        with self.assertRaises(TypeError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), blastdb_version='4')
        self.assertIn('must be an integer', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), blastdb_version=1)
        self.assertIn('supported values for `blastdb_version` are 4 and 5 only', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), blastdb_version=3)
        self.assertIn('supported values for `blastdb_version` are 4 and 5 only', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), blastdb_version=6)
        self.assertIn('supported values for `blastdb_version` are 4 and 5 only', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test.v4'), blastdb_version=4)
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test.v5'), blastdb_version=5)

    def test_max_file_sz(self):
        with self.assertRaises(TypeError) as cm:
            egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), max_file_sz=10)
        self.assertIn('must be a string', str(cm.exception))

        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test1'), max_file_sz="150MB")
        egglib.wrappers.makeblastdb(fname1, dbtype='nucl', out=self.tmpf('test2'), max_file_sz="2GB")

    def test_taxid(self):
        cnt = egglib.Container.create(alphabet=egglib.alphabets.protein, source=[
            ('seq1', 'MSFSTKPLDMATWPDFAALVERHNGVWGGCWCMAFHAKGSGAVGNREAKEARVREGSTHAALVFDGSACVGWCQFGPTGELPRIKHLRAYEDGQAVLPDWRITCFFSDKAFRGKGVAAAALAGALAEIGRLGGGTVESYPEDAQGRTVAGAFLHNGTLAM'),
            ('seq2', 'MKAIDLKAEEKKRLIEGIQDFFYEERNEEIGIIAAEKALDFFLSGVGKLIYNKALDESKIWFSRRLEDISLDYELLYK'),
            ('seq3', 'MTLAAAAQSATWTFIDGDWYEGNVAILGPRSHAMWLGTSVFDGARWFEGVAPDLELHAARVNASAIALGLAPNMTPEQIVGLTWDGLKKFDGKTAVYIRPMYWAEHGGYMGVPADPASTRFCLCLYESPMISPTGFSVTVSPFRRPTIETMPTNAKAGCLYPNNGRAILEAKARGFDNALVLDMLGNVAETGSSNIFLVKDGHVLTPAPNGTFLSGITRSRTMTLLGDYGFRTTEKTLSVRDFLEADEIFSTGNHSKVVPITRIEGRDLQPGPVAKKARELYWDWAHSASVG'),
            ('seq4', 'MRSFFHHVAAADPASFGVAQRVLTIPIKRAHIEVTHHLTKAEVDALIAAPNPRTSRGRRDRTFLLFLARTGARVSEATGVNANDLQLERSHPQVLLRGKGRRDRVIPIPQDLARALTALLAEHGIANHEPRPIFIGARQERLTRFGATHIVRRAAAQAVTIKPALAHKPISPHIFRHSLAMKLLQSGVDLLTIQAWLGHAQVATTHRYAAADVEMMRKGLEKAGVSGDLGLRFRPNDAVLQLLTSI'),
            ('seq5', 'MTISRVCGSRTEAMLTNGQEIAMTSILKSTGAVALLLLYTLTANATSLMISPSSIERVAPDRAAVFHLRNQMDRPISIKVRVFRWSQKGGVEKLEPTGDVVASPISAQLSPNGNRAVRVVRVSKEPLRSEEGYRVVIDEADPTRNTPEAESLSARHVLPVLFRPPDVLGPEIELSLTRSDGWLMLVVENKGASRLRRSDVTLAQGSAGIARREGFVGYVLPGLTRHWRVGREDSYSGGIVTVSANSSGGAIGEQLVVSGR'),
            ('seq6', 'TTLLLQVPIGWGVLHQGGALVVLGFAIAHWRGFVGTYTRDTAIEMRD')])

        f = open(self.tmpf('test_map.txt'), 'w')
        f.write("""lcl|seq1 68287
lcl|seq2 2382161
lcl|seq3 68287
lcl|seq4 382
lcl|seq5 382
lcl|seq6 382
""")
        f.close()

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(cnt, out=self.tmpf('test1'), taxid_map=self.tmpf('test_map.txt'))
        self.assertIn('`taxid_map` requires `parse_seqids`', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(cnt, out=self.tmpf('test1'), taxid_map=self.tmpf('prout.txt'), parse_seqids=True)
        self.assertIn('file passed for `taxid_map` not found', str(cm.exception))

        egglib.wrappers.makeblastdb(cnt, out=self.tmpf('test1'), taxid_map=self.tmpf('test_map.txt'), parse_seqids=True)
        egglib.wrappers.makeblastdb(self.tmpf('test1'), input_type='blastdb', dbtype= 'prot', out=self.tmpf('test2'), taxid_map=self.tmpf('test_map.txt'))

        with self.assertRaises(TypeError) as cm:
            egglib.wrappers.makeblastdb(cnt, parse_seqids=True, out=self.tmpf('test1'), taxid='382')
        self.assertIn('must be an integer', str(cm.exception))

        with self.assertRaises(TypeError) as cm:
            egglib.wrappers.makeblastdb(cnt, parse_seqids=True, out=self.tmpf('test1'), taxid=382.0)
        self.assertIn('`taxid` must be an integer', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(cnt, parse_seqids=True, out=self.tmpf('test1'), taxid=-1)
        self.assertIn('`taxid` must be >= 0', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            egglib.wrappers.makeblastdb(cnt, parse_seqids=True, out=self.tmpf('test1'), taxid_map=self.tmpf('test_map.txt'), taxid=382)
        self.assertIn('`taxid` and `taxid_map` are incompatible', str(cm.exception))

        egglib.wrappers.makeblastdb(cnt, out=self.tmpf('test3'), taxid=382)

class blastn_test(unittest.TestCase):
    def setUp(self):
        self.tmp = pathlib.Path(tempfile.mkdtemp())
        self.nucl = egglib.io.from_fasta(fname1, alphabet=egglib.alphabets.DNA)
        self.nucldb = str(self.tmp / 'nucl')
        egglib.wrappers.makeblastdb(self.nucl, out=self.nucldb)
        self.cnt = egglib.Container.create([
            ('name1', 'ACCGCGGAGCGCCGGAGAGCCCGCGGAGCTAGAGCTCTAGGAGCCT'),
            ('name2', 'ACGGTtgcccgcgaagcccgaagtgcgcgaGAGCGCCGGAGA'),
            ('name3', 'ATTGCAGTTGAGCCGCGAGTTGCGCATTGTTGCCGAGAGCGCGCCCGCGAAA'),
            ('name4', 'TTtgcccgcgaagcccgaagtgcgcgaAACCCGGAAAGtgcccgcgaagcccgaagtgcgcgaAG'),
            ('name5', 'TTGAGCCGAGAGCCGCGCGAGGAAGCGCCGCCGGTTGGCCGAGAGAGAAAGG')],
                    alphabet=egglib.alphabets.DNA)
        self.cntdb = str(self.tmp / 'cnt')
        egglib.wrappers.makeblastdb(self.cnt, out=self.cntdb)

        self.seq1 = 'ACCGCGGAGCGCCGGAGAGCCCGCGGAGCTAGAGCTCTAGGAGCCT'
        self.seq2 = 'ATTGCAGTTGAGCCGCGAGTTGCGCATTGTTGCCGAGAGCGCGCCCGCGAAA'
        self.query_align = egglib.Align.create([('q1', self.seq1[:40]), ('q2', self.seq2[:40])], alphabet=egglib.alphabets.DNA)
        self.query_cont = egglib.Container.create([('q1', self.seq1), ('q2', self.seq2)], alphabet=egglib.alphabets.DNA)
        self.init_path = pathlib.Path.cwd()

    def tearDown(self):
        os.chdir(self.init_path)
        shutil.rmtree(self.tmp)

    def test_blastn(self):
        res = egglib.wrappers.blastn(self.nucl[0], subject=self.nucl[0], parse_deflines=True,
                    query_loc=[0, 1000], subject_loc=[0, 500], evalue=1e-10,
                    num_threads=1, word_size=7, reward=2, penalty=-3, gapopen=4, gapextend=4,
                    no_dust=True, no_soft_masking=True)

        self.assertEqual(res.program, 'blastn')
        self.assertIsNone(res.db)
        self.assertEqual(res.query_ID, self.nucl.get_name(0))
        self.assertEqual(res.query_def, 'No definition line')
        self.assertEqual(res.query_len, 1000)
        self.assertEqual(res.params['expect'], 1e-10)
        self.assertEqual(res.params['sc-match'], 2)
        self.assertEqual(res.params['sc-mismatch'], -3)
        self.assertEqual(res.params['gap-open'], 4)
        self.assertEqual(res.params['gap-extend'], 4)
        self.assertEqual(res.params['filter'], 'F')
        self.assertNotIn('matrix', res.params)
        self.assertEqual(len(res), 1)
        [query] = res[:]
        self.assertEqual(query.num, 0)
        self.assertEqual(query.query_ID, self.nucl.get_name(0))
        self.assertEqual(query.query_def, 'No definition line')
        self.assertEqual(query.query_len, 1000)
        self.assertEqual(query.db_num, 0)
        self.assertEqual(query.db_len, 0)
        self.assertEqual(len(query), 1)
        [hit] = query[:]
        self.assertEqual(hit.num, 0)
        self.assertEqual(hit.id, self.nucl.get_name(0))
        self.assertEqual(hit.descr, 'No definition line')
        self.assertEqual(hit.accession, self.nucl.get_name(0))
        self.assertEqual(hit.len, self.nucl.get_sample(0).ls)
        self.assertEqual(len(hit), 1)
        [Hsp] = hit[:]
        self.assertEqual(Hsp.num, 0)
        self.assertLess(Hsp.evalue, 1e-20)
        self.assertEqual(Hsp.query_start, 0)
        self.assertEqual(Hsp.query_stop, 500)
        self.assertEqual(Hsp.query_frame, 1)
        self.assertEqual(Hsp.hit_start, 0)
        self.assertEqual(Hsp.hit_stop, 500)
        self.assertEqual(Hsp.hit_frame, 1)
        self.assertEqual(Hsp.identity, 500)
        self.assertEqual(Hsp.positive, 500)
        self.assertEqual(Hsp.gaps, 0)
        self.assertEqual(Hsp.align_len, 500)
        self.assertEqual(Hsp.qseq, self.nucl.get_sequence(0)[:500])
        self.assertEqual(Hsp.midline, '|' * 500)
        self.assertEqual(Hsp.hseq, self.nucl.get_sequence(0)[:500])

    def test_megablast_reverse(self):
        seq = egglib.tools.rc(self.nucl[0].sequence)
        ls = len(seq)
        res = egglib.wrappers.megablast(seq, subject=self.nucl[0], parse_deflines=True,
                    evalue=1e-10, num_threads=1, word_size=7,
                    reward=2, penalty=-3, gapopen=4, gapextend=4,
                    no_dust=False, no_soft_masking=False)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(len(res[0][0]), 1)
        Hsp = res[0][0][0]
        self.assertEqual(Hsp.query_start, 0)
        self.assertEqual(Hsp.query_stop, ls)
        self.assertEqual(Hsp.hit_start, 0)
        self.assertEqual(Hsp.hit_stop, ls)
        self.assertEqual(Hsp.query_frame, 1)
        self.assertEqual(Hsp.hit_frame, -1)

    def test_blastn_short(self):
        q = 'tgcccgcgaagcccgaagtgcgcga'
        res = egglib.wrappers.blastn_short(q, db=self.cntdb, evalue=1e-10)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 2)
        if res[0][0].descr == 'name2':
            hit1 = res[0][0]
            hit2 = res[0][1]
        else:
            hit1 = res[0][1]
            hit2 = res[0][0]
        self.assertEqual(hit1.descr, 'name2')
        self.assertEqual(hit2.descr, 'name4')
        self.assertEqual(len(hit1), 1)
        self.assertEqual(len(hit2), 2)
        Hsp1 = hit1[0]
        Hsp2 = hit2[0]
        Hsp3 = hit2[1]
        self.assertEqual(Hsp1.query_start, 0)
        self.assertEqual(Hsp1.query_stop, 25)
        self.assertEqual(Hsp1.hit_start, 5)
        self.assertEqual(Hsp1.hit_stop, 30)
        self.assertEqual(Hsp1.query_frame, 1)
        self.assertEqual(Hsp1.hit_frame, 1)
        self.assertEqual(Hsp2.query_start, 0)
        self.assertEqual(Hsp2.query_stop, 25)
        self.assertEqual(Hsp2.hit_start, 2)
        self.assertEqual(Hsp2.hit_stop, 27)
        self.assertEqual(Hsp2.query_frame, 1)
        self.assertEqual(Hsp2.hit_frame, 1)
        self.assertEqual(Hsp3.query_start, 0)
        self.assertEqual(Hsp3.query_stop, 25)
        self.assertEqual(Hsp3.hit_start, 38)
        self.assertEqual(Hsp3.hit_stop, 63)
        self.assertEqual(Hsp3.query_frame, 1)
        self.assertEqual(Hsp3.hit_frame, 1)
        for Hsp in [Hsp1, Hsp2, Hsp3]:
            self.assertEqual(Hsp.identity, 25)
            self.assertEqual(Hsp.positive, 25)
            self.assertEqual(Hsp.gaps, 0)
            self.assertEqual(Hsp.align_len, 25)
            self.assertEqual(Hsp.qseq, q.upper())
            self.assertEqual(Hsp.midline, '|' * 25)
            self.assertEqual(Hsp.hseq, q.upper())

    def test_query(self):
        for fun in [egglib.wrappers.blastn, egglib.wrappers.blastn_short,
                    egglib.wrappers.megablast, egglib.wrappers.dc_megablast]:

            # pass a string
            res = fun(query=self.seq1, db=self.cntdb, evalue=1e-15)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')

            # pass a SampleView
            res = fun(query=self.query_align[0], db=self.cntdb, evalue=1e-15)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')

            # pass a SequenceView
            res = fun(query=self.query_align[0].sequence, db=self.cntdb, evalue=1e-15)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')

            # pass an Align
            res = fun(query=self.query_align, db=self.cntdb, evalue=1e-15)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 2)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(len(res[1]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[1][0].descr, 'name3')

            # pass a Container
            res = fun(query=self.query_cont, db=self.cntdb, evalue=1e-10)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 2)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(len(res[1]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[1][0].descr, 'name3')

            # errors
            with self.assertRaises(TypeError):
                fun(query=list(self.seq1), db=self.cntdb)

    def test_subject(self):
        for fun in [egglib.wrappers.blastn, egglib.wrappers.blastn_short,
                    egglib.wrappers.megablast, egglib.wrappers.dc_megablast]:

            # pass a database
            res = fun(query=self.query_align, db=self.cntdb, evalue=1e-10)
            self.assertIsInstance(res, egglib.wrappers._blast.BlastOutput)
            self.assertEqual(len(res), 2)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(len(res[1]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[1][0].descr, 'name3')

            # cannot pass both
            with self.assertRaises(ValueError):
                fun(query=self.seq1, db=self.cntdb, subject=self.seq1, evalue=1e-10)

            # pass a string
            res = fun(query=self.seq1, subject=self.seq1)
            self.assertEqual(res[0][0][0].identity, len(self.seq1))

            # pass a SampleView
            res = fun(query=self.seq1, subject=self.query_cont[0])
            self.assertEqual(res[0][0][0].identity, len(self.seq1))

            # pass a SequenceView
            res = fun(query=self.seq1, subject=self.query_cont[0].sequence)
            self.assertEqual(res[0][0][0].identity, len(self.seq1))

    def test_loc(self):
        seq = self.nucl[0].sequence

        for fun in [egglib.wrappers.blastn, egglib.wrappers.blastn_short,
                    egglib.wrappers.megablast, egglib.wrappers.dc_megablast]:

            # use query_loc
            res = fun(seq, subject=seq, query_loc=(2000, 2500), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:2500])

            res = fun(seq, subject=seq, query_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            res = fun(self.nucl[0], subject=seq, query_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            res = fun(self.nucl[0].sequence, subject=seq, query_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            # invalid values
            with self.assertRaises(ValueError):
                res = fun(seq, subject=seq, query_loc=(-2000, 150000), evalue=1e-20)
            with self.assertRaises(ValueError):
                res = fun(seq, subject=seq, query_loc=(2000, 1000), evalue=1e-20)
            with self.assertRaises(ValueError):
                res = fun(self.nucl[0].sequence, subject=seq, query_loc=(200000, 300000), evalue=1e-20)

            # query must be a string
            with self.assertRaises(ValueError):
                res = fun(self.nucl, subject=seq, query_loc=(2000, 1000), evalue=1e-20)

            # use subject_loc
            res = fun(seq, subject=seq, subject_loc=(2000, 2500), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:2500])

            res = fun(seq, subject=seq, subject_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            res = fun(self.nucl[0], subject=seq, subject_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            res = fun(self.nucl[0].sequence, subject=seq, subject_loc=(2000, 150000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[2000:])

            res = fun(seq, subject=self.nucl[0], query_loc=(1000, 2500), subject_loc=(1500, 2000), evalue=1e-20)
            self.assertEqual(res[0][0][0].qseq, seq[1500:2000])

            # invalid values
            with self.assertRaises(ValueError):
                res = fun(seq, subject=seq, subject_loc=(-2000, 150000), evalue=1e-20)
            with self.assertRaises(ValueError):
                res = fun(seq, subject=seq, subject_loc=(2000, 1000), evalue=1e-20)
            with self.assertRaises(ValueError):
                res = fun(seq, subject=seq, subject_loc=(200000, 300000), evalue=1e-20)

            # subject must be a string
            with self.assertRaises(ValueError):
                res = fun(seq, subject=self.cntdb, subject_loc=(2000, 1000), evalue=1e-20)

    def test_evalue(self):
        res = egglib.wrappers.blastn_short(query=self.nucl[0], db=self.nucldb)
        self.assertEqual(res.params['expect'], 1000)
        for fun in [egglib.wrappers.blastn, egglib.wrappers.megablast,
                    egglib.wrappers.dc_megablast]:
            res = fun(query=self.nucl[0], db=self.nucldb)
            self.assertEqual(res.params['expect'], 10)

        res = egglib.wrappers.blastn_short(query=self.nucl[0], db=self.nucldb, evalue=1e-10)
        self.assertEqual(res.params['expect'], 1e-10)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastn_short(query=self.nucl[0], db=self.nucldb, evalue=0)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastn_short(query=self.nucl[0], db=self.nucldb, evalue=-1.0)

    def test_parse_deflines(self):
        for fun in [egglib.wrappers.blastn, egglib.wrappers.blastn_short,
                    egglib.wrappers.megablast, egglib.wrappers.dc_megablast]:
            res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, parse_deflines=True)
            self.assertEqual(res.query_ID, self.cnt[0].name)
            res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, parse_deflines=False)
            self.assertNotEqual(res.query_ID, self.cnt[0].name)

    def test_num_threads(self):
        for fun in [egglib.wrappers.blastn, egglib.wrappers.blastn_short,
                    egglib.wrappers.megablast, egglib.wrappers.dc_megablast]:
            res = fun(query=self.nucl[0], db=self.nucldb, num_threads=4)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastn_short(query=self.nucl[0], subject=self.nucl[0], num_threads=4)

    def test_word_size(self):
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, word_size=3)
        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, word_size=4)

        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, word_size=3)
        res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, word_size=4)

        with self.assertRaises(ValueError):
            res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, word_size=3)
        res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, word_size=4)

        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, word_size=4)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, word_size=10)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, word_size=13)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, word_size=28)
        res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, word_size=11)
        res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, word_size=12)

    def test_costs(self):
        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb)
        self.assertEqual(res.params['sc-match'], 2)
        self.assertEqual(res.params['sc-mismatch'], -3)
        self.assertEqual(res.params['gap-open'], 5)
        self.assertEqual(res.params['gap-extend'], 2)

        res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb)
        self.assertEqual(res.params['sc-match'], 1)
        self.assertEqual(res.params['sc-mismatch'], -3)
        self.assertEqual(res.params['gap-open'], 5)
        self.assertEqual(res.params['gap-extend'], 2)

        res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb)
        self.assertEqual(res.params['sc-match'], 1)
        self.assertEqual(res.params['sc-mismatch'], -2)

        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb)
        self.assertEqual(res.params['sc-match'], 2)
        self.assertEqual(res.params['sc-mismatch'], -3)
        self.assertEqual(res.params['gap-open'], 5)
        self.assertEqual(res.params['gap-extend'], 2)

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, reward=4, penalty=-5)
        self.assertEqual(res.params['sc-match'], 4)
        self.assertEqual(res.params['sc-mismatch'], -5)
        self.assertEqual(res.params['gap-open'], 12)
        self.assertEqual(res.params['gap-extend'], 8)

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb,
                    reward=1, penalty=-3, gapopen=2, gapextend=2)
        self.assertEqual(res.params['sc-match'], 1)
        self.assertEqual(res.params['sc-mismatch'], -3)
        self.assertEqual(res.params['gap-open'], 2)
        self.assertEqual(res.params['gap-extend'], 2)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb,
                    reward=1, penalty=-3, gapopen=3, gapextend=2)

    def test_strand(self):
        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='both', evalue=1e-20)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')
        self.assertEqual(res[0][0][0].query_frame, 1)
        self.assertEqual(res[0][0][0].hit_frame, 1)

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='plus', evalue=1e-20)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')
        self.assertEqual(res[0][0][0].query_frame, 1)
        self.assertEqual(res[0][0][0].hit_frame, 1)

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='minus', evalue=1e-20)
        self.assertEqual(len(res[0]), 0)

        res = egglib.wrappers.blastn(query=egglib.tools.rc(self.cnt[0].sequence), db=self.cntdb, strand='both', evalue=1e-20)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')
        self.assertEqual(res[0][0][0].query_frame, 1)
        self.assertEqual(res[0][0][0].hit_frame, -1)

        res = egglib.wrappers.blastn(query=egglib.tools.rc(self.cnt[0].sequence), db=self.cntdb, strand='plus', evalue=1e-20)
        self.assertEqual(len(res[0]), 0)

        res = egglib.wrappers.blastn(query=egglib.tools.rc(self.cnt[0].sequence), db=self.cntdb, strand='minus', evalue=1e-20)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')
        self.assertEqual(res[0][0][0].query_frame, 1)
        self.assertEqual(res[0][0][0].hit_frame, -1)

    def test_filters(self):
        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='both', evalue=1e-20)
        self.assertEqual(res.params['filter'], 'L;m;')

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='both', evalue=1e-20, no_dust=True)
        self.assertEqual(res.params['filter'], 'm;')

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='both', evalue=1e-20, no_soft_masking=True)
        self.assertEqual(res.params['filter'], 'L;')

        res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, strand='both', evalue=1e-20, lcase_masking=True)
        self.assertEqual(res.params['filter'], 'L;m;')

        s1 = 'GTGCTACACCTCAACCCGGTAGGCTGACAGGGGCTCGGCCTCgccgttttaaatcctgagtacatccttagtctgcatttcgcctaggtgcccgcagcattgggctagGTGGGCGTATTGACTAAGAATTACAGATTGGGAAATTGAGAATCACATGGAGGTAGTGGAGAAACTAAAAAGCGGGTAATATCGGTGTAGTT'
        s2 = 'GACTAAGAGTGCCAAGGGAGCTCGTAGGAATTCAGAACAAGATTTACCAGGTATATGCATACCCGTGAAATGGGCCATATCTTCAGATTGAGAGATTGTCTGCCTCTTCAATATCCCACTATCTCATCTAGCCGTTTTAAATCCTGAGTACATCCTTAGTCTGCATTTCGCCTAGGTGCCCGCAGCATTGGGCTAGAACG'
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20)
        self.assertEqual(len(res[0]), 1)
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20, lcase_masking=True)
        self.assertEqual(len(res[0]), 0)

    def test_perc_identity(self):
        s1 = 'GACTAAGAGTGCCAAGGGAGCTCGTAGGAATTCAGAACAAGATTTACCAGGTATATGCATACCCGTGAAATGGGCCATATCTTCAGATTGAGAGATTGTCTGCCTCTTCAATATCCCACTATCTCATCTAGCCGTTTTAAATCCTGAGTACATCCTTAGTCTGCA'
        s2 = 'GACTAAGAGTGCCAAGGGAGCTCGTAGGAATTCAGAACAAGATTTACCAGGTATATGCATACCCGTGAAATGGGCCATATCTTCAGATTGAGAGATTGTCTGCCTCTTCAATATCCCACTATCTCATCTAGCCGTTTTAAATCCTGAGTACATCCTTAGTCTGCA'
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0][0].query_start, 0)
        self.assertEqual(res[0][0][0].query_stop, len(s1))
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20, perc_identity=100)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0][0].query_start, 0)
        self.assertEqual(res[0][0][0].query_stop, len(s1))
        #                                                                    C                                                                         T
        s1 = 'GACTAAGAGTGCCAAGGGAGCTCGTAGGAATTCAGAACAAGATTTACCAGGTATATGCATACCGGTGAAATGGGCCATATCTTCAGATTGAGAGATTGTCTGCCTCTTCAATATCCCACTATCTCATCTAGCCGTTTAAAATCCTGAGTACATCCTTAGTCTGCA'
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0][0].query_start, 0)
        self.assertEqual(res[0][0][0].query_stop, len(s1))
        res = egglib.wrappers.blastn(query=s1, subject=s2, evalue=1e-20, perc_identity=100)
        self.assertEqual(len(res[0]), 0)

    def test_template_params(self):
        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=18)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=18)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=18)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=18)
        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=16)
        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=21)
        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='optimal', template_length=18)
        res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding_and_optimal', template_length=18)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='xxx', template_length=18)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=15)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=17)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=19)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=20)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=22)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, template_type='coding', template_length=0)

    def test_greedy(self):
        res = egglib.wrappers.megablast(query=self.cnt[0], db=self.cntdb, no_greedy=True)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.blastn(query=self.cnt[0], db=self.cntdb, no_greedy=True)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.blastn_short(query=self.cnt[0], db=self.cntdb, no_greedy=True)
        with self.assertRaises(TypeError):
            res = egglib.wrappers.dc_megablast(query=self.cnt[0], db=self.cntdb, no_greedy=True)

    def test_relative_path(self):
        os.chdir(self.tmp)
        res = egglib.wrappers.blastn(query=self.seq1, db='cnt', evalue=1e-15)
        self.assertIsInstance(res, egglib.wrappers.BlastOutput)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')

        os.chdir('..')
        res = egglib.wrappers.blastn(query=self.seq1, db=str(self.tmp / 'cnt'), evalue=1e-15)
        self.assertIsInstance(res, egglib.wrappers.BlastOutput)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')
        
        os.chdir(self.tmp)
        pathlib.Path('test_relative_path_subdir').mkdir()
        os.chdir('test_relative_path_subdir')
        res = egglib.wrappers.blastn(query=self.seq1, db=str(pathlib.Path('..') / 'cnt'), evalue=1e-15)
        self.assertIsInstance(res, egglib.wrappers.BlastOutput)
        self.assertEqual(len(res), 1)
        self.assertEqual(len(res[0]), 1)
        self.assertEqual(res[0][0].descr, 'name1')


class blastp_test(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.prot = egglib.io.from_fasta(fname2, alphabet=egglib.alphabets.protein)
        self.protdb = os.path.join(self.tmp, 'prot')
        egglib.wrappers.makeblastdb(self.prot, out=self.protdb)
 
    def tearDown(self):
        shutil.rmtree(self.tmp)

    def test_blastp(self):
        res = egglib.wrappers.blastp(self.prot[0], subject=self.prot[0],
                    parse_deflines=True, evalue=1e-5, num_threads=1, word_size=7)

        self.assertEqual(res.program, 'blastp')
        self.assertIsNone(res.db)
        self.assertEqual(res.query_ID, self.prot.get_name(0).split()[0])
        self.assertEqual(res.query_def, ' '.join(self.prot.get_name(0).split()[1:]))
        self.assertEqual(res.query_len, self.prot[0].ls)
        self.assertEqual(res.params['expect'], 1e-5)
        self.assertNotIn('sc-match', res.params)
        self.assertNotIn('sc-mismatch', res.params)
        self.assertEqual(res.params['gap-open'], 11)
        self.assertEqual(res.params['gap-extend'], 1)
        self.assertEqual(res.params['filter'], 'F')
        self.assertEqual(len(res), 1)
        [query] = res[:]
        self.assertEqual(query.num, 0)
        self.assertEqual(query.query_ID, self.prot.get_name(0).split()[0])
        self.assertEqual(query.query_def, ' '.join(self.prot.get_name(0).split()[1:]))
        self.assertEqual(query.query_len, self.prot[0].ls)
        self.assertEqual(query.db_num, 0)
        self.assertEqual(query.db_len, 0)
        self.assertEqual(len(query), 1)
        [hit] = query[:]
        self.assertEqual(hit.num, 0)
        self.assertEqual(hit.id, self.prot.get_name(0).split()[0])
        self.assertEqual(hit.descr, ' '.join(self.prot.get_name(0).split()[1:]))
        self.assertEqual(hit.accession, self.prot.get_name(0).split()[0])
        self.assertEqual(hit.len, self.prot[0].ls)
        self.assertEqual(len(hit), 1)
        [Hsp] = hit[:]
        self.assertEqual(Hsp.num, 0)
        self.assertLess(Hsp.evalue, 1e-20)
        self.assertEqual(Hsp.query_start, 0)
        self.assertEqual(Hsp.query_stop, self.prot[0].ls)
        self.assertEqual(Hsp.query_frame, 0)
        self.assertEqual(Hsp.hit_start, 0)
        self.assertEqual(Hsp.hit_stop, self.prot[0].ls)
        self.assertEqual(Hsp.hit_frame, 0)
        self.assertEqual(Hsp.identity, self.prot[0].ls)
        self.assertEqual(Hsp.positive, self.prot[0].ls)
        self.assertEqual(Hsp.gaps, 0)
        self.assertEqual(Hsp.align_len, self.prot[0].ls)
        self.assertEqual(Hsp.qseq, self.prot.get_sequence(0).string())
        self.assertEqual(Hsp.midline, self.prot.get_sequence(0).string())
        self.assertEqual(Hsp.hseq, self.prot.get_sequence(0).string())

    def test_word_size(self):
        for fun in egglib.wrappers.blastp, egglib.wrappers.blastp_short, egglib.wrappers.blastp_fast:
            with self.assertRaises(ValueError):
                fun(self.prot[0], db=self.protdb, word_size=0)
            with self.assertRaises(ValueError):
                fun(self.prot[0], db=self.protdb, word_size=1)
            fun(self.prot[0], db=self.protdb, word_size=2)
            fun(self.prot[0], db=self.protdb, word_size=3)
            fun(self.prot[0], db=self.protdb, word_size=4)
            fun(self.prot[0], db=self.protdb, word_size=5)
            fun(self.prot[0], db=self.protdb, word_size=6)
            fun(self.prot[0], db=self.protdb, word_size=7)
            with self.assertRaises(ValueError):
                fun(self.prot[0], db=self.protdb, word_size=8)

    def test_gap_costs(self):
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30', gapopen=0, gapextend=0)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30', gapopen=8, gapextend=2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30', gapopen=32767, gapextend=32767)
        self.assertEqual(res.params['gap-open'], 32767)
        self.assertEqual(res.params['gap-extend'], 32767)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30', gapopen=10, gapextend=1)
        self.assertEqual(res.params['gap-open'], 10)
        self.assertEqual(res.params['gap-extend'], 1)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30', gapopen=8, gapextend=1)
        self.assertEqual(res.params['gap-open'], 8)
        self.assertEqual(res.params['gap-extend'], 1)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30')
        self.assertEqual(res.params['gap-open'], 9)
        self.assertEqual(res.params['gap-extend'], 1)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70', gapopen=0, gapextend=0)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70', gapopen=10, gapextend=2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70', gapopen=32767, gapextend=32767)
        self.assertEqual(res.params['gap-open'], 32767)
        self.assertEqual(res.params['gap-extend'], 32767)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70', gapopen=8, gapextend=2)
        self.assertEqual(res.params['gap-open'], 8)
        self.assertEqual(res.params['gap-extend'], 2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70', gapopen=12, gapextend=3)
        self.assertEqual(res.params['gap-open'], 12)
        self.assertEqual(res.params['gap-extend'], 3)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70')
        self.assertEqual(res.params['gap-open'], 10)
        self.assertEqual(res.params['gap-extend'], 1)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62', gapopen=0, gapextend=0)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62', gapopen=8, gapextend=1)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62', gapopen=32767, gapextend=32767)
        self.assertEqual(res.params['gap-open'], 32767)
        self.assertEqual(res.params['gap-extend'], 32767)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62', gapopen=8, gapextend=2)
        self.assertEqual(res.params['gap-open'], 8)
        self.assertEqual(res.params['gap-extend'], 2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62', gapopen=11, gapextend=1)
        self.assertEqual(res.params['gap-open'], 11)
        self.assertEqual(res.params['gap-extend'], 1)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62')
        self.assertEqual(res.params['gap-open'], 11)
        self.assertEqual(res.params['gap-extend'], 1)

        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=0, gapextend=0)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=10, gapextend=2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=32767, gapextend=32767)
        self.assertEqual(res.params['gap-open'], 32767)
        self.assertEqual(res.params['gap-extend'], 32767)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=8, gapextend=2)
        self.assertEqual(res.params['gap-open'], 8)
        self.assertEqual(res.params['gap-extend'], 2)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=11, gapextend=1)
        self.assertEqual(res.params['gap-open'], 11)
        self.assertEqual(res.params['gap-extend'], 1)
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80', gapopen=10, gapextend=1)
        self.assertEqual(res.params['gap-open'], 10)
        self.assertEqual(res.params['gap-extend'], 1)

    def test_matrix(self):
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM62')
        self.assertEqual(res.params['matrix'], 'BLOSUM62')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='BLOSUM80')
        self.assertEqual(res.params['matrix'], 'BLOSUM80')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM30')
        self.assertEqual(res.params['matrix'], 'PAM30')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='PAM70')
        self.assertEqual(res.params['matrix'], 'PAM70')
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, matrix='prout')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb)
        self.assertEqual(res.params['matrix'], 'BLOSUM62')
        res = egglib.wrappers.blastp_short(self.prot[0], db=self.protdb)
        self.assertEqual(res.params['matrix'], 'PAM30')

    def test_threshold(self):
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=2)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=10)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=100)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=0)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=-1)
        with self.assertRaises(TypeError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, threshold=0.5)

    def test_comp_based_stats(self):
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=0)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=1)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=2)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=3)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=-1)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, comp_based_stats=4)

    def test_seg(self):
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=0)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=1)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(12, 2.2, 2.5))
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=-1)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=3)
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(0, 2.2, 2.5))
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(1, 2.2, 2.5))
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(10, 0, 2.5))
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(10, 0.1, 2.5))
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(10, 2.2, 2.1))
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, seg=(10, 1, 10))

    def test_filters(self):
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb)
        self.assertEqual(res.params['filter'], 'F')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, lcase_masking=True)
        self.assertEqual(res.params['filter'], 'F')
        res = egglib.wrappers.blastp(self.prot[0], db=self.protdb, soft_masking=True)
        self.assertEqual(res.params['filter'], 'm;')

    def test_sw(self):
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, use_sw_tback=True)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, use_sw_tback=False)

    def test_window_size(self):
        with self.assertRaises(ValueError):
            egglib.wrappers.blastp(self.prot[0], db=self.protdb, window_size=-1)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, window_size=0)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, window_size=50)
        egglib.wrappers.blastp(self.prot[0], db=self.protdb, window_size=1000)

class blastx_test(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.cds = egglib.io.from_fasta(fname3, alphabet=egglib.alphabets.DNA)
        self.prot = egglib.tools.translate(egglib.tools.to_codons(self.cds))
        self.protdb = os.path.join(self.tmp, 'prot')
        egglib.wrappers.makeblastdb(self.prot, out=self.protdb)
 
    def tearDown(self):
        shutil.rmtree(self.tmp)

    def test_blastx(self):
        ls = self.cds[0].ls

        for f in egglib.wrappers.blastx, egglib.wrappers.blastx_fast:
            res = f(self.cds[0], subject=self.prot[0],
                        evalue=1e-5, num_threads=1, word_size=None, seg=0)

            self.assertEqual(res.program, 'blastx')
            self.assertIsNone(res.db)
            self.assertEqual(res.query_ID, 'Query_1')
            self.assertEqual(res.query_def, 'lcl|' + self.cds.get_name(0))
            self.assertEqual(res.query_len, ls)
            self.assertEqual(res.params['expect'], 1e-5)
            self.assertEqual(res.params['filter'], 'F')
            self.assertEqual(len(res), 1)
            [query] = res[:]
            self.assertEqual(query.num, 0)
            self.assertEqual(query.query_ID, 'Query_1')
            self.assertEqual(query.query_def, 'lcl|' + self.cds.get_name(0))
            self.assertEqual(query.query_len, ls)
            self.assertEqual(query.db_num, 0)
            self.assertEqual(query.db_len, 0)
            self.assertEqual(len(query), 1)
            [hit] = query[:]
            self.assertEqual(hit.num, 0)
            self.assertEqual(hit.id, 'lcl|' + self.cds.get_name(0))
            self.assertEqual(hit.descr, 'lcl|' + self.cds.get_name(0))
            self.assertEqual(hit.accession, 'Subject_1')
            self.assertEqual(hit.len, ls//3)
            self.assertEqual(len(hit), 1)
            [Hsp] = hit[:]
            self.assertEqual(Hsp.num, 0)
            self.assertLess(Hsp.evalue, 1e-20)
            self.assertEqual(Hsp.query_start, 0)
            self.assertEqual(Hsp.query_stop, ls)
            self.assertEqual(Hsp.query_frame, 1)
            self.assertEqual(Hsp.hit_start, 0)
            self.assertEqual(Hsp.hit_stop, ls//3)
            self.assertEqual(Hsp.hit_frame, 0)
            self.assertEqual(Hsp.identity, ls//3)
            self.assertEqual(Hsp.positive, ls//3)
            self.assertEqual(Hsp.gaps, 0)
            self.assertEqual(Hsp.align_len, ls//3)
            self.assertEqual(Hsp.qseq, self.prot.get_sequence(0).string())
            self.assertEqual(Hsp.midline, self.prot.get_sequence(0).string())
            self.assertEqual(Hsp.hseq, self.prot.get_sequence(0).string())

    def test_strand(self):
        for f in egglib.wrappers.blastx, egglib.wrappers.blastx_fast:
            res = f(query=self.cds[0], db=self.protdb, strand='both', evalue=1e-20)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[0][0][0].query_frame, 1)
            self.assertEqual(res[0][0][0].hit_frame, 0)

            res = f(query=self.cds[0], db=self.protdb, strand='plus', evalue=1e-20)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[0][0][0].query_frame, 1)
            self.assertEqual(res[0][0][0].hit_frame, 0)

            res = f(query=self.cds[0], db=self.protdb, strand='minus', evalue=1e-20)
            self.assertEqual(len(res[0]), 0)

            res = f(query=egglib.tools.rc(self.cds[0].sequence), db=self.protdb, strand='both', evalue=1e-20)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[0][0][0].query_frame, -1)
            self.assertEqual(res[0][0][0].hit_frame, 0)

            res = f(query=egglib.tools.rc(self.cds[0].sequence), db=self.protdb, strand='plus', evalue=1e-20)
            self.assertEqual(len(res[0]), 0)

            res = f(query=egglib.tools.rc(self.cds[0].sequence), db=self.protdb, strand='minus', evalue=1e-20)
            self.assertEqual(len(res), 1)
            self.assertEqual(len(res[0]), 1)
            self.assertEqual(res[0][0].descr, 'name1')
            self.assertEqual(res[0][0][0].query_frame, -1)
            self.assertEqual(res[0][0][0].hit_frame, 0)

    def test_genetic_code(self):
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=0)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=1)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=2)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=3)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=4)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=5)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=6)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=7)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=8)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=9)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=10)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=12)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=15)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=16)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=17)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=19)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=20)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=21)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=22)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=23)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=24)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=25)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=25)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=26)
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, query_genetic_code=30)

    def test_max_intron_length(self):
        with self.assertRaises(TypeError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, max_intron_length='a')
        with self.assertRaises(ValueError):
            res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, max_intron_length=-1)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, max_intron_length=0)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, max_intron_length=20)
        res = egglib.wrappers.blastx(query=self.cds[0], db=self.protdb, max_intron_length=5000)

class tblastn_test(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.cds = egglib.io.from_fasta(fname3, alphabet=egglib.alphabets.DNA)
        self.prot = egglib.tools.translate(egglib.tools.to_codons(self.cds))
        self.cdsdb = os.path.join(self.tmp, 'cds')
        egglib.wrappers.makeblastdb(self.cds, out=self.cdsdb)
 
    def tearDown(self):
        shutil.rmtree(self.tmp)

    def test_tblastn(self):
        ls = self.prot[0].ls

        for f in egglib.wrappers.tblastn, egglib.wrappers.tblastn_fast:
            res = f(self.prot[0], subject=self.cds[0],
                        evalue=1e-5, num_threads=1, word_size=None, seg=0)

            self.assertEqual(res.program, 'tblastn')
            self.assertIsNone(res.db)
            self.assertEqual(res.query_ID, 'Query_1')
            self.assertEqual(res.query_def, 'lcl|' + self.prot.get_name(0))
            self.assertEqual(res.query_len, ls)
            self.assertEqual(res.params['expect'], 1e-5)
            self.assertEqual(res.params['filter'], 'F')
            self.assertEqual(len(res), 1)
            [query] = res[:]
            self.assertEqual(query.num, 0)
            self.assertEqual(query.query_ID, 'Query_1')
            self.assertEqual(query.query_def, 'lcl|' + self.prot.get_name(0))
            self.assertEqual(query.query_len, ls)
            self.assertEqual(query.db_num, 0)
            self.assertEqual(query.db_len, 0)
            self.assertEqual(len(query), 1)
            [hit] = query[:]
            self.assertEqual(hit.num, 0)
            self.assertEqual(hit.id, 'lcl|' + self.prot.get_name(0))
            self.assertEqual(hit.descr, 'lcl|' + self.prot.get_name(0))
            self.assertEqual(hit.accession, 'Subject_1')
            self.assertEqual(hit.len, ls*3)
            self.assertEqual(len(hit), 1)
            [Hsp] = hit[:]
            self.assertEqual(Hsp.num, 0)
            self.assertLess(Hsp.evalue, 1e-20)
            self.assertEqual(Hsp.query_start, 0)
            self.assertEqual(Hsp.query_stop, ls)
            self.assertEqual(Hsp.query_frame, 0)
            self.assertEqual(Hsp.hit_start, 0)
            self.assertEqual(Hsp.hit_stop, ls*3)
            self.assertEqual(Hsp.hit_frame, 1)
            self.assertEqual(Hsp.identity, ls)
            self.assertEqual(Hsp.positive, ls)
            self.assertEqual(Hsp.gaps, 0)
            self.assertEqual(Hsp.align_len, ls)
            self.assertEqual(Hsp.qseq, self.prot.get_sequence(0).string())
            self.assertEqual(Hsp.midline, self.prot.get_sequence(0).string())
            self.assertEqual(Hsp.hseq, self.prot.get_sequence(0).string())

    def test_soft_masking(self):
        res = egglib.wrappers.tblastn(self.prot[0], subject=self.cds[0],
                    evalue=1e-5, num_threads=1, word_size=None, seg=0)
        self.assertEqual(res.params['filter'], 'F')

        res = egglib.wrappers.tblastn(self.prot[0], subject=self.cds[0],
                    evalue=1e-5, num_threads=1, word_size=None, seg=0,
                    soft_masking=True)
        self.assertEqual(res.params['filter'], 'm;')

    def test_db_genetic_code(self):
        for i in [1,2,3,4,5,6,9,10,11,12,13,14,15,16,21,22,23,24,25]:
            egglib.wrappers.tblastn(self.prot[0], subject=self.cds[0],
                db_genetic_code=i)
        for i in [-1, 0, 7, 8, 17, 18, 19, 20, 26, 27, 28, 30, 50, 100]:
            with self.assertRaises(ValueError):
                egglib.wrappers.tblastn(self.prot[0], subject=self.cds[0],
                    db_genetic_code=i)

    def test_max_intron_length(self):
        with self.assertRaises(TypeError):
            egglib.wrappers.tblastn(query=self.prot[0], db=self.cdsdb, max_intron_length='a')
        with self.assertRaises(ValueError):
            egglib.wrappers.tblastn(query=self.prot[0], db=self.cdsdb, max_intron_length=-1)
        egglib.wrappers.tblastn(query=self.prot[0], db=self.cdsdb, max_intron_length=0)
        egglib.wrappers.tblastn(query=self.prot[0], db=self.cdsdb, max_intron_length=20)
        egglib.wrappers.tblastn(query=self.prot[0], db=self.cdsdb, max_intron_length=5000)

class tblastx_test(unittest.TestCase):
    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        self.cds = egglib.io.from_fasta(fname3, alphabet=egglib.alphabets.DNA)
        self.cdsdb = os.path.join(self.tmp, 'cds')
        egglib.wrappers.makeblastdb(self.cds, out=self.cdsdb)
 
    def tearDown(self):
        shutil.rmtree(self.tmp)

    def test_tblastx(self):
        ls = self.cds[0].ls

        res = egglib.wrappers.tblastx(self.cds[0], subject=self.cds[0],
                    evalue=1e-20, num_threads=1, word_size=None)

        self.assertEqual(res.program, 'tblastx')
        self.assertIsNone(res.db)
        self.assertEqual(res.query_ID, 'Query_1')
        self.assertEqual(res.query_def, 'lcl|' + self.cds.get_name(0))
        self.assertEqual(res.query_len, ls)
        self.assertEqual(res.params['expect'], 1e-20)
        self.assertEqual(res.params['filter'], 'L;')
        self.assertEqual(len(res), 1)
        [query] = res[:]
        self.assertEqual(query.num, 0)
        self.assertEqual(query.query_ID, 'Query_1')
        self.assertEqual(query.query_def, 'lcl|' + self.cds.get_name(0))
        self.assertEqual(query.query_len, ls)
        self.assertEqual(query.db_num, 0)
        self.assertEqual(query.db_len, 0)
        self.assertEqual(len(query), 1)
        [hit] = query[:]
        self.assertEqual(hit.num, 0)
        self.assertEqual(hit.id, 'lcl|' + self.cds.get_name(0))
        self.assertEqual(hit.descr, 'lcl|' + self.cds.get_name(0))
        self.assertEqual(hit.accession, 'Subject_1')
        self.assertEqual(hit.len, ls)
        self.assertEqual(len(hit), 6)
        self.assertEqual([Hsp.num for Hsp in hit], list(range(6)))
        for Hsp in hit: self.assertLess(Hsp.evalue, 1e-20)
        self.assertEqual(set([Hsp.query_start for Hsp in hit]), set([0, 1, 2]))
        self.assertEqual(set([Hsp.query_stop for Hsp in hit]), set([ls-2, ls-1, ls]))
        self.assertEqual(set([Hsp.query_frame for Hsp in hit]), set([-3, -2, -1, 1, 2, 3]))
        self.assertEqual(set([Hsp.hit_start for Hsp in hit]), set([0, 1, 2]))
        self.assertEqual(set([Hsp.hit_stop for Hsp in hit]), set([ls-2, ls-1, ls]))
        self.assertEqual(set([Hsp.hit_frame for Hsp in hit]), set([-3, -2, -1, 1, 2, 3]))
        self.assertEqual(set([Hsp.identity for Hsp in hit]), set([ls//3, ls//3-1]))
        self.assertEqual(set([Hsp.positive for Hsp in hit]), set([ls//3, ls//3-1]))
        self.assertEqual(set([Hsp.gaps for Hsp in hit]), set([0]))
        self.assertEqual(set([Hsp.align_len for Hsp in hit]), set([ls//3, ls//3-1]))
        self.assertEqual(hit[0].qseq, egglib.tools.translate(self.cds.get_sequence(0).string()))
        self.assertEqual(hit[0].midline, egglib.tools.translate(self.cds.get_sequence(0).string()))
        self.assertEqual(hit[0].hseq, egglib.tools.translate(self.cds.get_sequence(0).string()))

    def test_soft_masking(self):
        res = egglib.wrappers.tblastx(self.cds[0], subject=self.cds[0],
                    evalue=1e-5, num_threads=1, word_size=None)
        self.assertEqual(res.params['filter'], 'L;')

        res = egglib.wrappers.tblastx(self.cds[0], subject=self.cds[0],
                    evalue=1e-5, num_threads=1, word_size=None,
                    soft_masking=True)
        self.assertEqual(res.params['filter'], 'L;m;')

    def test_db_genetic_code(self):
        for i in [1,2,3,4,5,6,9,10,11,12,13,14,15,16,21,22,23,24,25]:
            egglib.wrappers.tblastx(self.cds[0], subject=self.cds[0],
                db_genetic_code=i)
        for i in [-1, 0, 7, 8, 17, 18, 19, 20, 26, 27, 28, 30, 50, 100]:
            with self.assertRaises(ValueError):
                egglib.wrappers.tblastx(self.cds[0], subject=self.cds[0],
                    db_genetic_code=i)



"""
    Copyright 2023 Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, re
from .. import _interface, alphabets

def from_genepop(fname):
    """
    Import Genepop-formatted genotypic data. The format is described
    `here <https://genepop.curtin.edu.au/help_input.html>`_.

    :param fname: Genepop-formatted file name.

    :return: A new :class:`.Align` instance.

    The returned object contains data mapped to an *ad hoc* alphabet,
    with two samples per individuals. Group labels are used to indicate
    the structure (first level: populations, second level: individuals).
    In addition to normal :class:`.Align` instance, the returned object
    has two attributes: :attr:`!title` and :attr:`!loci`, which contain
    the information read from the Genepop file.
    """

    with open(fname) as f:
        # read title
        title = f.readline().strip()
        if len(title) == 0: raise ValueError('title cannot be empty')

        # read locus names
        loci = []
        for line in f:
            if line == '': raise ValueError('unexpected end of line')
            line = line.strip()
            if line in {'POP', 'Pop', 'pop'}: break
            loci.extend(map(str.strip, re.split(', ?', line)))

        # initialise data
        res = _interface.Align(alphabets.genepop)

        # read pops
        pop_idx = 0
        while True:
            for line in f:
                line = line.strip()
                if line in {'POP', 'Pop', 'pop'}:
                    pop_idx += 1
                    break

                # read indiv
                hit = re.match('(.+),(.+)', line)
                if hit is None: raise ValueError('invalid line: {0}'.format(line))
                name, genos = hit.groups()
                name = name.strip()
                genos = list(map(str.strip, genos.split()))
                if len(genos) != len(loci): raise ValueError('inconsistent number of loci')

                # read genotypes and import
                lens = set(map(len, genos))
                if lens == {4}:
                    res.add_sample(name+'_1', [int(geno[:2]) for geno in genos], ['pop{0}'.format(pop_idx+1), name])
                    res.add_sample(name+'_2', [int(geno[2:]) for geno in genos], ['pop{0}'.format(pop_idx+1), name])
                elif lens == {6}:
                    res.add_sample(name+'_1', [int(geno[:3]) for geno in genos], ['pop{0}'.format(pop_idx+1), name])
                    res.add_sample(name+'_2', [int(geno[3:]) for geno in genos], ['pop{0}'.format(pop_idx+1), name])
                else:
                    raise ValueError('invalid number of characters in genotype')
            else:
                break

        # return alignment
        res.title = title
        res.loci = loci
        return res



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class Phyml_test(unittest.TestCase):

    def test_Phyml_T(self):
        tree = egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')

        aln = egglib.io.from_fasta(str(path / 'example.fas'), egglib.alphabets.DNA, labels=True)
        tree, stats = egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)
        self.assertIsInstance(tree, egglib.Tree)
        self.assertIsInstance(stats, dict)

        aln.to_codons()
        prot = egglib.tools.translate(aln)
        tree, stats = egglib.wrappers.phyml(prot, model='LG', verbose=False,
                start_tree='pars',
                alpha=None, rates=4, pinv=0.3, use_median=False, freq='m',
                free_rates=False)
        self.assertIsInstance(tree, egglib.Tree)
        self.assertIsInstance(stats, dict)

    def test_Phyml_E(self):
        tree = egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset:0.02254897,Tamarin:0.01990484):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')

        cache = egglib.wrappers.paths['phyml']
        egglib.wrappers.paths['phyml'] = None
        aln = egglib.io.from_fasta(str(path / 'example.fas'), egglib.alphabets.DNA, labels=True)
        with self.assertRaises(RuntimeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)
        egglib.wrappers.paths.get('phyml').set_path_force(cache)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml('aln', model='TN93', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        alne0 = egglib.io.from_fasta(str(path / 'cds_e.fas'), egglib.alphabets.DNA, labels=True)
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(alne0, model='TN93', verbose=False,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        alne1 = egglib.io.from_fasta(str(path / 'example_Els.fas'), egglib.alphabets.DNA, labels=True)
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(alne1, model='TN93', verbose=False,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model=1500, verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='eerroorr', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='FAaIiL', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    boot='error', start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    boot=-10, start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, TiTv=-150 ,  rates=2, pinv=None, freq='m',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='error',
                    use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='ACTG',
                    use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=2, pinv=150, freq='m',
                use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates='ERROR', pinv=None, freq='m',
                use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=0, pinv=None, freq='m',
                use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha='ERROR', rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=-150.0, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)


        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree,
                    fixed_brlens=True,  alpha=None, rates=1, pinv=None, freq='m',
                use_median=True, free_rates=True)


        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree='tree',
                    fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)

        alne2 = egglib.io.from_fasta(str(path / 'example_edpl.fas'), egglib.alphabets.DNA, labels=True)
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(alne2, model='TN93', verbose=False,
                start_tree=tree,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)

        tree0 = egglib.Tree(string='(Hello:0.01257025,Iam:0.02023601,(An:0.03625789,((Phyml:0.02002846,Error:0.02646824):0.00119894):0.01464801):0.01145931);')
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree0,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)

        aln = egglib.io.from_fasta(str(path / 'example_Els.fas'), egglib.alphabets.DNA, labels=True)	
        tree1 = egglib.Tree(string='(Hello:0.01257025,Iam:0.02023601,(An:0.03625789,((Phyml:0.02002846,Error:0.02646824):0.00119894):0.01464801):0.01145931);')
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree1,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)
    
        aln = egglib.io.from_fasta(str(path / 'example.fas'), egglib.alphabets.DNA, labels=True)	
        tree2 = egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789):0.01145931);')
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree2,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)

        tree4 = egglib.Tree(string='(Spider:0.01257025,Woolly:0.02023601,(Howler:0.03625789,((Titi:0.02002846,Saki:0.02646824):0.01312676,((Owl:0.02467454,(((Gorilla:0.00570008,(Human:0.00467442,Chimp:0.00218595):0.00198277):0.00803251,(Gibbon:0.02031871,Orangutan:0.01428695):0.00052501):0.01497214,(Colobus:0.00134948,(DLangur:0.00479108,(Patas:0.01038680,((AGM_cDNA:0.00067736,Tant_cDNA:0.00000006):0.00511480,(Baboon:0.00531400,Rhes_cDNA:0.00519922):0.00433750):0.00200495):0.00621549):0.00134574):0.02864316):0.11226732):0.00500924,(Squirrel:0.04657578,(PMarmoset,Tamarin):0.01687811):0.00000008):0.00119894):0.01464801):0.01145931);')
        aln = egglib.io.from_fasta(str(path / 'example.fas'), egglib.alphabets.DNA, labels=True)	
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                start_tree=tree4,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None, freq='m',
                use_median=True, free_rates=True)
        
        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                fixed_brlens=True,  alpha=None, rates=2, pinv=None,
                freq='m', use_median=True, free_rates=True)

        with self.assertRaises(TypeError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    fixed_brlens=False,  alpha=None, 
                rates=2, pinv=None, freq='m', use_median=True, free_rates=True, seed='error')

        with self.assertRaises(ValueError):
            egglib.wrappers.phyml(aln, model='TN93', verbose=False,
                    start_tree=tree, fixed_brlens=True,  alpha=None, 
                rates=2, pinv=None, freq='m', use_median=True, free_rates=True, seed=-10)

    def test_alphabet(self):
        dna = egglib.Align.create(
               [('one',   'GGCGAGCGACGCCCCTGGCGA'),
                ('two',   'GGCGAGAGACTCCTCTGGCGA'),
                ('three', 'GGCGTGAGACGGCTCTGCCTA'),
                ('four',  'GGCGTGAGACGGCACTGCCTA')], egglib.alphabets.DNA)
        tree, stats = egglib.wrappers.phyml(dna, 'HKY85')
        with self.assertRaises(ValueError):
            tree, stats = egglib.wrappers.phyml(dna, 'JTT')

        cod = egglib.Align.create(
               [('one',   ['GGC','GAG','CGA','CGC','CCC','TGG','CGA']),
                ('two',   ['GGC','GAG','AGA','CTC','CTC','TGG','CGA']),
                ('three', ['GGC','GTG','AGA','CGG','CTC','TGA','CTA']),
                ('four',  ['GGC','GTG','AGA','CGG','CAC','TGA','CTA'])], egglib.alphabets.codons)
        tree, stats = egglib.wrappers.phyml(cod, 'HKY85')
        with self.assertRaises(ValueError):
            tree, stats = egglib.wrappers.phyml(cod, 'JTT')

        dna.to_codons()
        prot = egglib.tools.translate(dna)
        tree, stats = egglib.wrappers.phyml(prot, 'JTT')
        with self.assertRaises(ValueError):
            tree, stats = egglib.wrappers.phyml(prot, 'GTR')

        char = egglib.Align.create(
               [('one',   'GGCGAGCGACGCCCCTGGCGA'),
                ('two',   'GGCGAGAGACTCCTCTGGCGA'),
                ('three', 'GGCGTGAGACGGCTCTGACTA'),
                ('four',  'GGCGTGAGACGGCACTGACTA')], egglib.alphabets.Alphabet('char', 'ACGT', None))
        with self.assertRaises(ValueError):
            tree, stats = egglib.wrappers.phyml(char, 'HKY85')

    def test_labels(self):
        cs = egglib.coalesce.Simulator(2, num_chrom=[4, 4], migr=0.01, theta=4, num_alleles=4)
        aln = cs.simul()
        dna = egglib.Align(egglib.alphabets.DNA)
        for i, seq in enumerate(aln):
            dna.add_sample('seq{0}'.format(i+1), ['ACGT'[i] for i in seq.sequence], seq.labels)

        tree, lk = egglib.wrappers.phyml(dna, 'HKY85', labels=False, verbose=False)
        T = egglib.Tree(string=tree.newick())
        self.assertListEqual(sorted([n.label for n in T.iter_leaves()]), sorted(dna.names()))

        tree, lk = egglib.wrappers.phyml(dna, 'HKY85', labels=True)
        ctrl = [seq.name + '@' + ','.join(seq.labels) for seq in dna]
        self.assertListEqual(sorted([n.label for n in tree.iter_leaves()]), sorted(ctrl))
        tree.newick()



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class EHH_test(unittest.TestCase):
    def test_EHH_T(self):
        ehh=egglib.stats.EHH()
        self.assertEqual(str(type(ehh)), "<class 'egglib.stats._ehh.EHH'>")

    def test_set_core_T(self): # unfinished test 
        ehh=egglib.stats.EHH()
        aln = egglib.io.from_fasta(str(path / 'dmi3.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        site = egglib.site_from_align(aln, 2213)
        site.position = 0
        ehh.set_core(site)
        sites = []
        alph = egglib.alphabets.Alphabet('range', (1, 99), (0, ))
        idx = 1
        for site in (([1,1, 1,0, 1,1, 1,1, 2,1, 1,2, 2,0, 2,2, 2,2, 2,2, 0,0, 3,3, 3,3]),
                     ([1,0, 0,0, 1,2, 1,1, 2,1, 2,2, 2,0, 2,2, 2,2, 2,2, 0,0, 3,3, 1,3]),
                     ([4,2, 4,3, 3,1, 3,2, 1,2, 1,4, 1,2, 3,2, 1,3, 2,4, 3,1, 2,2, 3,1]),
                     ([4,2, 4,3, 3,1, 3,2, 1,1, 3,4, 1,2, 3,2, 1,3, 2,4, 3,1, 2,2, 3,1]),
                     ([4,2, 4,3, 3,1, 3,2, 2,2, 1,1, 1,2, 3,2, 1,3, 2,4, 3,1, 2,2, 3,1])):
            site = egglib.site_from_list(site, alphabet=alph)
            site.position = idx
            sites.append(site)
            idx += 1
        ehh.set_core(sites[0], struct=2, EHHS_thr=0.5, min_freq=None)
        for site in sites[1:]: ehh.load_distant(site)

    def test_structure(self):
        core =  '00000111111111122222222223'
        dist1 = '00000000111111111111111111'
        dist2 = '00010110000011111101100000'
        dist3 = '01110000000111000000111100'
        dist4 = '00111111011101111111111111'
        site_core = egglib.site_from_list(core, alphabet=egglib.alphabets.Alphabet('char', '0123', []))
        site_core.position = 0
        alph = egglib.alphabets.Alphabet('char', '01', [])
        site1 = egglib.site_from_list(dist1, alphabet=alph)
        site2 = egglib.site_from_list(dist2, alphabet=alph)
        site3 = egglib.site_from_list(dist3, alphabet=alph)
        site4 = egglib.site_from_list(dist4, alphabet=alph)
        site1.position = 1
        site2.position = 2
        site3.position = 3
        site4.position = 4

        ehh = egglib.stats.EHH()

        ehh.set_core(site_core)
        self.assertAlmostEqual(ehh.num_haplotypes, 4)
        self.assertAlmostEqual(ehh.get_EHH(0), 20/20.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(1), 90/90.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(2), 90/90.0, places=6)
        self.assertIsNone(ehh.get_EHH(3))

        ehh.load_distant(site1)
        self.assertAlmostEqual(ehh.get_EHH(0), 20/20.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(1), 48/90.0, places=6) 
        self.assertAlmostEqual(ehh.get_EHH(2), 90/90.0, places=6)
        self.assertIsNone(ehh.get_EHH(3))

        ehh.load_distant(site2)
        self.assertAlmostEqual(ehh.get_EHH(0), 12/20.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(1), 20/90.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(2), 40/90.0, places=6)
        self.assertIsNone(ehh.get_EHH(3))

        ehh.load_distant(site3)
        self.assertAlmostEqual(ehh.get_EHH(0), 4/20.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(1), 10/90.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(2), 20/90.0, places=6)
        self.assertIsNone(ehh.get_EHH(3))

        ehh.load_distant(site4)
        self.assertAlmostEqual(ehh.get_EHH(0), 0/20.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(1), 4/90.0, places=6)
        self.assertAlmostEqual(ehh.get_EHH(2), 20/90.0, places=6)
        self.assertIsNone(ehh.get_EHH(3))

        struct = egglib.struct_from_dict(
            {None: {None: {'0': (0, 1), '1': (2, 3), '2': (4, 5), '3': (6, 7), '4': (8, 9), '5': (10, 11),  '6': (12, 13),
                     '7': (14, 15), '8': (16, 17), '9': (18, 19), '10': (20, 21), '11': (22, 23), '12': (24, 25)}}}, None)

        ehh.set_core(site_core, struct=struct)
        self.assertEqual(ehh.get_EHHG(), 1.0)
        ehh.load_distant(site1)
        self.assertEqual(ehh.get_EHHG(), 1.0)
        ehh.load_distant(site2)
        self.assertEqual(ehh.get_EHHG(), 0.6)
        ehh.load_distant(site3)
        self.assertEqual(ehh.get_EHHG(), 0.4)
        ehh.load_distant(site4)
        self.assertEqual(ehh.get_EHHG(), 0.2)

        ehh.set_core(site_core, struct=2)
        self.assertEqual(ehh.get_EHHG(), 1.0)
        ehh.load_distant(site1)
        self.assertEqual(ehh.get_EHHG(), 1.0)
        ehh.load_distant(site2)
        self.assertEqual(ehh.get_EHHG(), 0.6)
        ehh.load_distant(site3)
        self.assertEqual(ehh.get_EHHG(), 0.4)
        ehh.load_distant(site4)
        self.assertEqual(ehh.get_EHHG(), 0.2)



"""
    Copyright 2016-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, re, tempfile, shutil, pathlib, platformdirs
from .. import _interface

#### write sequences down a file #######################################

def _write_sample(obj, fname):
    if not isinstance(obj, (_interface.SampleView, str)): raise TypeError('invalid type for `obj` argument')
    f = open(fname, 'w')

    if isinstance(obj, _interface.SampleView):
        name = obj[0]
        sequence =list(obj[1].string())
        f.write('>{0}\n'.format(name))
        for j in range(0, len(sequence), 60):
            f.write(''.join(sequence[j:j+60]) + '\n')
    elif isinstance(obj, str):
        sequence =list(obj)
        f.write('>String_to_fasta\n')
        for j in range(0, len(sequence), 60):
            f.write(''.join(sequence[j:j+60]) + '\n')
    f.close()

def _write(obj, fname, mapping):
    f = open(fname, 'w')
    for i, sam in enumerate(obj):
        n = 'seq-{0}'.format(len(mapping) + 1)
        f.write('>{0}\n'.format(n))
        for j in range(0, len(sam.sequence), 100):
            f.write(sam.sequence[j:j+100] + '\n')
        mapping[n] = sam
    f.close()

#### decorator running a function in a temp directory ##################

_protect_path_mapping = {}
def _protect_run(f):
    """
    This function is designed to be a decorator for wrapper functions.
    It runs the argument function into a temporary directory, ensuring
    that the temporary directory is deleted after completion of the
    function.
    """
    def _f(*args, **kwargs):
        tmp = None
        curr = os.getcwd()
        try:
            tmp = tempfile.mkdtemp()
            _protect_path_mapping[tmp] = curr
            os.chdir(tmp)
            return f(*args, **kwargs)
        finally:
            os.chdir(curr)
            if tmp is not None:
                shutil.rmtree(tmp)
            del _protect_path_mapping[tmp]
    _f.__name__ = f.__name__
    _f.__doc__ = f.__doc__
    return _f

#### external application paths manager ################################

class _Paths(object):
    """
    Class (designed to be a singleton) holding external applications.
    Each application is represented by an _App instance. Instances are
    iterable (iteration over application keys instances). Paths can be
    set/accessed through the [] operator.
    """

    def __init__(self):
        self._fname_g = pathlib.Path(__file__).parent / 'apps.conf'
        self._fname_u = platformdirs.user_config_path(appname='EggLib') / 'apps.conf'

        if self._fname_u.is_file():
            self._fname = self._fname_u
        else:
            if not self._fname_g.is_file(): raise RuntimeError('apps.conf file not found')
            self._fname = self._fname_g
        self._apps = {}

    @property
    def fname(self):
        """
        Location of the configuration file.
        """
        return self._fname

    def _add(self, app):
        """
        Add an application wrapper.

        :param app: a :class:`._App` instance.
        :param key: the application's name.
        :param default: the default path.
        """
        self._apps[app._key] = app

    def __iter__(self):
        for i in self._apps:
            yield i

    def __getitem__(self, app):
        if app not in self._apps: raise ValueError('invalid application name: {0}'.format(app))
        return self._apps[app].get_path()

    def get(self, app, default=None):
        """
        Get the _App object corresponding to an application key. If key
        doesn't exist return default value otherwise raise a KeyError.
        """
        try: return self._apps[app]
        except KeyError: return default

    def __setitem__(self, app, path):
        if app not in self._apps: raise ValueError('invalid application name: {0}'.format(app))
        self._apps[app].set_path(path, True)

    def autodetect(self, verbose=False):
        """
        Try to set all paths using default values. Applications who are
        currently configured are skipped.
        """
        failed = {}
        passed = 0
        skipped = 0
        n = len(self._apps)
        sz = len(str(n))
        if verbose:
            print('Detecting external applications: ' + '0'.rjust(sz) + '/' + str(n), flush=True, end='')
        for key, app in self._apps.items():
            if app.get_path() is not None:
                skipped += 1
                if verbose:
                    print('\b'*(2*sz+1) + str(passed+skipped+len(failed)).rjust(sz) + '/' + str(n), flush=True, end='')
                continue
            result = app.set_path(app._default, False)
            if result is None:
                passed += 1
            else:
                failed[key] = app._default, result
            if verbose:
                print('\b'*(2*sz+1) + str(passed+skipped+len(failed)).rjust(sz) + '/' + str(n), flush=True, end='')
        if verbose:
            print('\b'*(2*sz+1) + str(passed).rjust(sz) + '/' + str(n) + f' passed, {skipped} skipped')
            for k, (cmd, msg) in failed.items():
                print(k, ' [', cmd, ']: ', msg, sep='')
        return passed, len(failed), failed

    def save(self, dest=None):
        """
        Save current configuration to configuration file (requires
        administrator rights if the file is located in a protected
        location).

        :param dest: if 'global', force save to global path (within the
            EggLib distribution). If 'user', force save as user config
            file. All other values (default), same as when loading.
        """
        if dest == 'global':
            self._fname = self._fname_g
        elif dest == 'user':
            self._fname_u.parent.mkdir(parents=True, exist_ok=True)
            self._fname = self._fname_u
        try:
            f = open(self.fname, 'w')
        except IOError as e:
            if e.errno == 13: raise ValueError('administrator rights are required')
            else: raise
        for app in self._apps.values(): app.save(f)
        f.close()

    def load(self):
        """
        Load configuration from the persistent file.
        """
        f = open(self.fname)
        for linenum, line in enumerate(f):
            line = line.strip()
            if line == '': continue         # support empty lines
            if line.lstrip()[0] == '#': continue     # support comment lines
            if (mo := re.fullmatch('([a-z]+?):[ \t]*(.+)\s*', line)) is None:
                raise RuntimeError('invalid file `apps.conf` (line {0})'.format(linenum+1))
            key, rvalue = mo.groups()
            if key not in self._apps: raise RuntimeError('invalid file `apps.conf` (unknown application: {0})'.format(key))
            self._apps[key].load(rvalue, linenum)
        f.close()

    def __str__(self):
        return str(dict([(i, j.get_path()) for (i,j) in self._apps.items()]))

    def as_dict(self):
        """
        Return configuration as a new dict instance.
        """
        return dict([(i, j.get_path()) for (i,j) in self._apps.items()])

    def clear(self):
        """
        Set all paths to None
        """
        for p in self._apps.values():  p.zero()

    def delete_user(self):
        """
        Delete user configuration. The installation will fall back to
        the default configuration file present in EggLib's installation
        directory.
        """
        self._fname_u.unlink(missing_ok=True)

class _App(object):
    def __init__(self, key, default):
        self._path = None
        self._key = key
        self._default = default
        self.config = {} # option values (only if path is specified)
        self.options = {}

    def add_option(self, **args):
        self.options[args['name']] = args
        self.config[args['name']] = args['default']

    def get_path(self):
        """ Get the application path. """
        return self._path

    def set_path(self, path, critical):
        """
        Set the application path. The path is set if the check function
        (to be implemented as _check_path) succeeds. If it fails, a
        :exc:`~.exceptions.ValueError` is raised (if *critical* is
        ``True``), or the path is set to ``None`` and the error message
        string is returned (otherwise).
        """
        if path is None:
            self._path = None
            self.config = { opt['name']: opt['default'] for opt in self.options.values() }
        else:
            result = self._check_path(path, self.config) # _check_path is responsible of setting options
            if result is None: self._path = path
            elif critical: raise ValueError('cannot set path for {0}: {1}'.format(self._key, result))
            else: return result

    def _set_options(self, options, key):
        if (diff := set(options) - set(self.options)) != set(): raise ValueError('unexpected option(s): {",".join(diff)}')
        for o in set(self.options) - set(options):
            self.config[o] = self.options[o][key]()
        for o, v in options.items():
            try:
                self.config[o] = self.options[o]['f_convert'](v)
            except ValueError:
                raise ValueError(f'invalid option value for option {o}: {v}')

    def set_path_force(self, path, **options):
        """
        Set path without performing the check function.
        """
        self._set_options(options, 'f_set_missing')
        self._path = path

    def save(self, f):
        """
        Write app configuration in file-like object.
        """
        if self._path is None:
            f.write(f'{self._key}: *\n')
        else:
            f.write('{0}: {1}'.format(self._key, self._path))
            if len(self.config) == 0: f.write('\n')
            else:
                for k, v in self.config.items(): f.write(f' ${k}={v}')
                f.write('\n')

    def load(self, string, num):
        """
        Import app configuration from string.
        """
        if (mo := re.fullmatch('((?![ \$]).+?)[ \t]*(\$.+)?', string)) is None:
            raise RuntimeError('invalid file `apps.conf` (line {0})'.format(num+1))
        path, options = mo.groups()
        if path == '*':
            self._path = None
        else:
            self._path = path
            if options is not None: options = {k: v for k, v in re.findall('\$(\w+)=(\w+)', options)}
            else: options = {}
            self._set_options(options, 'f_load_missing')

    def zero(self):
        """
        Set path to None
        """
        self._path = None
        for k in self.config:
            self.config[k] = self.options[k]['default']

    key = property(lambda self: self._key, doc='key of the object _App')

paths = _Paths()



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, random, time
from random import *
from scipy.stats import bernoulli, binom, expon, geom, ttest_ind, zmap, ttest_ind, norm, poisson, binom_test as binom_testf # prevent loading binom_test as a test class
import numpy as np

class Random_test(unittest.TestCase):
    def test_bernoulli_T(self):
        size = 10000
        res = [egglib.random.bernoulli(0.5) for i in range(size)]
        self.assertSetEqual(set(res), {True, False})
        self.assertGreater(res.count(True), 200)
        self.assertGreater(res.count(False), 200)
        self.assertListEqual([egglib.random.bernoulli(0) for i in range(size)], [False] * size)
        self.assertListEqual([egglib.random.bernoulli(1) for i in range(size)], [True] * size)

    def test_bernoulli_E(self):
        with self.assertRaises(ValueError):
            egglib.random.bernoulli(-10)
        with self.assertRaises(ValueError):
            egglib.random.bernoulli(10)

    def test_binomial_T(self):

        for rep in range(10):
            distribution1=[]
            distribution2=[]
            k=2
            for i in range(1000):
                rb=binom.rvs(6, 0.5, size=20).tolist()
                distribution1.append(rb.count(k))
                rb2=[egglib.random.binomial(6, 0.5) for i in range(20)]
                distribution2.append(rb2.count(k))
            T=ttest_ind(distribution1, distribution2)
            if T.pvalue > 0.05:
                break
        else:
            self.assertTrue(False, 'test of binomial distribution failed too many times')

    def test_binomial_E(self):
        with self.assertRaises(ValueError):
            egglib.random.binomial(-10,0.2)
        with self.assertRaises(ValueError):
            egglib.random.binomial(10, -0.2)
        with self.assertRaises(ValueError):
            egglib.random.binomial(10, 10)

    def test_exponential_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range(1000):
                re=[egglib.random.exponential(1/0.2) for i in range(20)]
                distribution1.append(sum(re)/20)
                re2=expon.rvs(scale=(1/0.2), size=20)
                distribution2.append(sum(re2)/20)
            #T-ttest
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0

    def test_exponential_E(self):
        with self.assertRaises(ValueError):
            egglib.random.exponential(-10)

    def test_geometric_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range(1000):
                rg=[egglib.random.geometric(0.2) for i in range(10)]
                rg2=geom.rvs(0.2, size=10)
                distribution1.append(sum(rg)/10)
                distribution2.append(sum(rg2)/10)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05) 
                j=1
            except:
                j=0

    def test_geometrie_E(self):
        with self.assertRaises(ValueError):
            egglib.random.geometric(-10)
        with self.assertRaises(ValueError):
            egglib.random.geometric(10)

    def test_normal_T(self):
        #source: http://wwwf.imperial.ac.uk/~naheard/C245/hypothesis_testing_article.pdf
        j=0
        while (j==0): 
            distribution1=[egglib.random.normal() for i in range(1000)]
            distribution2 = norm.rvs(size=1000)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1
            except: 
                j=0

        j=0
        while (j==0): 
            distribution1=[egglib.random.normal_bounded(1, 2, -1000, +1000) for i in range(1000)]
            distribution2 = norm.rvs(loc=1, scale=2, size=1000)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1
            except: 
                j=0


        for i in range(100):
            x = egglib.random.normal_bounded(1, 1, 0.9, 3)
            self.assertGreaterEqual(x, 0.9)
            self.assertLessEqual(x, 3)


    def test_poisson_T(self):
        distribution1=[egglib.random.poisson(0.2) for i in range(1000)]
        distribution2=[]
        for i in range (1000):
            r=poisson.rvs(0.2)
            distribution2.append(r)

        nzero1=distribution1.count(0)
        ntrue1=1000-nzero1
        nzero2=distribution2.count(0)
        ntrue2=1000-nzero2

        #C-test
        x=ntrue1
        n=ntrue1+ntrue2
        p=len(distribution1)/(len(distribution1)+len(distribution2))
        C= binom_testf(x,n,p)
        self.assertTrue(C<1.96)

    def test_poisson_E(self):
        with self.assertRaises(ValueError):
            egglib.random.poisson(0)
        with self.assertRaises(ValueError):
            egglib.random.poisson(-10)

    def test_integer_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.integer(20)
                distribution1.append(ri1)
                ri2=np.random.randint(20,size=1)  # , dtype='int' [disabled because not available before numpy 1.11 ; int is the default]
                distribution2.append(ri2[0])

            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0

    def test_integer_E(self):
        with self.assertRaises(ValueError):
            egglib.random.integer(0)
        with self.assertRaises(ValueError):
            egglib.random.integer(-10)

    def test_integer_32bit_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.integer_32bit()
                distribution1.append(ri1)
                ri2=getrandbits(32)
                distribution2.append(ri2)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0
        
    def test_uniform_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.uniform()
                distribution1.append(ri1)
                ri2=uniform(0,1)
                distribution2.append(ri2)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0

    def test_uniform_53bit_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.uniform_53bit()
                distribution1.append(ri1)
                ri2=uniform(0,1) #Random, produces 53-bit precision floats and has a period of 2**19937-1.
                distribution2.append(ri2)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0
            
    def test_uniform_closed_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.uniform_closed()
                distribution1.append(ri1)
                ri2=np.float32(uniform(0.0000000, 1.0000001))
                distribution2.append(ri2)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0

    def test_uniform_open_T(self):
        distribution1=[]
        distribution2=[]
        j=0
        while (j==0):
            for i in range (1000):
                ri1=egglib.random.uniform_open()
                distribution1.append(ri1)
                ri2=np.float32(uniform(0.0000001, 1.0000000))
                distribution2.append(ri2)
            T=ttest_ind(distribution1, distribution2)
            try:
                self.assertTrue(T[1]>0.05)
                j=1 
            except:
                j=0

    def test_get_seed_T(self):
        time.sleep(1)
        egglib.random.set_seed(1500000000)
        self.assertEqual(egglib.random.get_seed(),1500000000)

    def test_set_seed_T(self):
        time.sleep(1)
        egglib.random.set_seed(1500000000)
        self.assertEqual(egglib.random.get_seed(),1500000000)

    def test_set_seed_E(self):
        with self.assertRaises(TypeError):
            egglib.random.set_seed(100.002)

    def test_seed(self):
        egglib.random.set_seed(593291034)
        X1 = tuple(egglib.random.uniform() for i in range(10))
        egglib.random.set_seed(593291034)
        X2 = tuple(egglib.random.uniform() for i in range(10))
        egglib.random.set_seed(593291034)
        X3 = tuple(egglib.random.uniform() for i in range(10))
        egglib.random.set_seed(593291034)
        X4 = tuple(egglib.random.uniform() for i in range(10))
        egglib.random.set_seed(593291034)
        X5 = tuple(egglib.random.uniform() for i in range(10))

        self.assertEqual(len(set([X1, X2, X3, X4, X5])), 1)

        X6 = tuple(egglib.random.uniform() for i in range(10))
        X7 = tuple(egglib.random.uniform() for i in range(10))
        X8 = tuple(egglib.random.uniform() for i in range(10))
        X9 = tuple(egglib.random.uniform() for i in range(10))
        Xa = tuple(egglib.random.uniform() for i in range(10))

        self.assertEqual(len(set([X5, X6, X7, X8, X9, Xa])), 6)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class Paralog_pi_test(unittest.TestCase):
    def test_ParalogPi_T(self):
        prlpi=egglib.stats.ParalogPi()
        self.assertIsInstance(prlpi, egglib.stats.ParalogPi)
        
    def test_setup_T(self):
        prlpi=egglib.stats.ParalogPi()
        ns = 15
        sim = egglib.coalesce.Simulator(4, num_indiv=[ns]*4, migr=1.0, theta=8.0)
        paralogs_aln = sim.simul()
        pstruct = egglib.struct_from_labels(paralogs_aln, lvl_pop=0)
        istruct = {}
        for i in range(ns*2):
                istruct[str(i)] = dict([(str(i+j*ns*2), (i+j*ns*2,)) for j in range(4)])
        istruct = egglib.struct_from_dict({None: istruct}, None)
        prlpi.setup(pstruct, istruct)
        self.assertEqual(prlpi.num_sites(),0)
        self.assertEqual(pstruct.req_ns,120)
        self.assertEqual(istruct.req_ns,120)
        self.assertIsInstance(prlpi, egglib.stats.ParalogPi)


    def test_process_align_T(self):
        prlpi=egglib.stats.ParalogPi()
        ns = 15
        sim = egglib.coalesce.Simulator(4, num_indiv=[ns]*4, migr=1.0, theta=8.0)
        paralogs_aln = sim.simul()
        pstruct = egglib.struct_from_labels(paralogs_aln, lvl_pop=0)
        istruct = {}
        for i in range(ns*2):
                istruct[str(i)] = dict([(str(i+j*ns*2), (i+j*ns*2,)) for j in range(4)])
        istruct = egglib.struct_from_dict({None: istruct}, None)
        prlpi.setup(pstruct, istruct)
        num_sites_b=prlpi.num_sites(0)
        self.assertIsInstance(prlpi, egglib.stats.ParalogPi)
        prlpi.process_align(paralogs_aln)
        num_sites_a=prlpi.num_sites(0)
        self.assertTrue(num_sites_a>num_sites_b)
    
    
    def test_process_align_E(self):
        prlpi=egglib.stats.ParalogPi()
        ns = 15
        sim = egglib.coalesce.Simulator(4, num_indiv=[ns]*4, migr=1.0, theta=8.0)
        paralogs_aln = sim.simul()
        pstruct = egglib.struct_from_labels(paralogs_aln, lvl_pop=0)
        aln = egglib.io.from_fasta(str(path / '125_seqs_300_300_bp.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        istruct = {}
        for i in range(ns*2):
                istruct[str(i)] = dict([(str(i+j*ns*2), (i+j*ns*2,)) for j in range(4)])
        istruct = egglib.struct_from_dict({None: istruct}, None)
        prlpi.setup(pstruct, istruct)
        with self.assertRaises(ValueError):
            prlpi.process_align(aln)


    def test_process_site_T(self):
        prlpi=egglib.stats.ParalogPi()
        ns = 50
        sim = egglib.coalesce.Simulator(4, num_indiv=[ns]*4, migr=1.0, theta=8.0)
        paralogs_aln = sim.simul()
        pstruct = egglib.struct_from_labels(paralogs_aln, lvl_pop=0)
        istruct = {}
        for i in range(ns*2):
                istruct[str(i)] = dict([(str(i+j*ns*2), (i+j*ns*2,)) for j in range(4)])
        istruct = egglib.struct_from_dict({None: istruct}, None)
        prlpi.setup(pstruct, istruct)
        num_sites_b=prlpi.num_sites(0)
        self.assertIsInstance(prlpi, egglib.stats.ParalogPi)

        num_sites_a=prlpi.num_sites(0)
        aln=egglib.io.from_fasta(str(path / '50_seqs_100_100_bp.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        sites_aln=egglib.Site()
        for i in range(aln.ls): sites_aln.from_align(aln, i)
        sites_aln.ns
        sites_aln.num_missing
        sites_aln.as_list()

        #prlpi.process_site(sites_aln)
        #sites_aln.process_align(aln,10)
        #num_sites_b=prlpi.num_sites(0)
        #prlpi.process_site(sites_aln)

        #num_sites_a=prlpi.num_sites(0)
        #self.assertTrue(num_sites_a>num_sites_b)
    
    def test_process_site_E(self):
        prlpi=egglib.stats.ParalogPi()
        ns = 15
        sim = egglib.coalesce.Simulator(4, num_indiv=[ns]*4, migr=1.0, theta=8.0)
        paralogs_aln = sim.simul()
        pstruct = egglib.struct_from_labels(paralogs_aln, lvl_pop=0)
        istruct = {}
        for i in range(ns*2):
                istruct[str(i)] = dict([(str(i+j*ns*2), (i+j*ns*2,)) for j in range(4)])
        istruct = egglib.struct_from_dict({None: istruct}, None)
        prlpi.setup(pstruct, istruct)
        num_sites_b=prlpi.num_sites(0)
        self.assertIsInstance(prlpi, egglib.stats.ParalogPi)

        num_sites_a=prlpi.num_sites(0)
        aln=egglib.io.from_fasta(str(path / '125_seqs_300_300_bp.fas'), labels=True, alphabet=egglib.alphabets.DNA)
        sites_aln=egglib.Site()
        for i in range(aln.ns): sites_aln.from_align(aln, i)
        with self.assertRaises(ValueError):
            prlpi.process_site(sites_aln)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, sys, unittest, random, re, gc, time, tempfile, pathlib
import collections
path = pathlib.Path(__file__).parent / '..' / 'data'

class Fasta_test(unittest.TestCase):
    def test_from_fasta_T(self):
        fname_T='align1_T.fas'
        aln = egglib.io.from_fasta(os.path.join(path, fname_T), egglib.alphabets.DNA)
        self.assertIsInstance(aln, egglib.Align)
        self.assertEqual(aln.ns, 101)
        self.assertEqual(aln.ls, 8942)

    def test_from_fasta_E(self):
        fname_T='align1_T.fas'
        fname_F='align1_F.fas'
        with self.assertRaises(ValueError):
            aln = egglib.io.from_fasta(os.path.join(path, fname_T), egglib.alphabets.DNA, labels=False, cls='LOL') 
            #an error integrated in the "cls" parameter
        with self.assertRaises(ValueError):
            aln = egglib.io.from_fasta(os.path.join(path, fname_F), egglib.alphabets.DNA, labels=False, cls='Align') 
            #The file 'align1_E.fs', contains sequences of differents sizes

    def setUp(self):
        fname='align1_T.fas'
        self.iter_aln=egglib.io.fasta_iter(os.path.join(path, fname), egglib.alphabets.DNA)

    def tearDown(self):
        del self.iter_aln

    def test_fasta_iter_T(self):
        self.assertIsInstance(self.iter_aln, egglib.io.fasta_iter)
        self.assertIsInstance(self.iter_aln, collections.abc.Iterable)

    def test__enter__T(self):
        with self.iter_aln as iter_aln:
            self.assertIsInstance(iter_aln, egglib.io.fasta_iter)

    def test__exit__T(self):
        ext=self.iter_aln.__exit__(None, None, None)
        self.assertFalse(ext)
        with self.assertRaises(StopIteration):
            next(self.iter_aln)

    def test__iter__T(self):
        self.assertIsInstance(self.iter_aln, collections.abc.Iterable)

    def test_next_T(self):
        self.assertIsInstance(next(self.iter_aln), egglib._interface.SampleView)

    def test_next_E(self):
        fname_E='align1_E.fas'
        iter_alnE=egglib.io.fasta_iter(os.path.join(path, fname_E), egglib.alphabets.DNA, True)
        with self.assertRaises(StopIteration):
            next(iter_alnE)

    def test_separator_E(self):
        try:
            f, fname = tempfile.mkstemp()
            os.write(f, b'''\
>one@a,a
AAAAAAAAAAAAAAAAAAAA
>two@a,a
AAAAAAAAAAAAAAAAAAAA
>three@a,b
AAAAAAAAAAAAAAAAAAAA
>four@a,b
AAAAAAAAAAAAAAAAAAAA
>five@b,c
AAAAAAAAAAAAAAAAAAAA
>six@b,c
AAAAAAAAAAAAAAAAAAAA
>seven@b,d
AAAAAAAAAAAAAAAAAAAA
>eight@b,d
AAAAAAAAAAAAAAAAAAAA
>outgroup@#
AAAAAAAAAAAAAAAAAAAA
''')
            os.close(f)
            aln = egglib.io.from_fasta(fname, egglib.alphabets.DNA, True)
            aln2 = egglib.io.from_fasta(fname, egglib.alphabets.DNA, True)

            self.assertListEqual([list(sam[2]) for sam in aln], [list(sam[2]) for sam in aln2])

            aln2[0].labels[1] = 'a-x'
            aln2[1].labels[1] = 'a-x'

            aln2.fasta(fname, labels=True)
            aln3 = egglib.io.from_fasta(fname, egglib.alphabets.DNA, True)
            self.assertListEqual([list(sam[2]) for sam in aln2], [list(sam[2]) for sam in aln3])

            aln2[0].labels[1] = 'a,x'
            aln2[1].labels[1] = 'a,x'

            with self.assertRaises(ValueError):
                aln2.fasta(fname, labels=True)
        finally:
            if os.path.isfile(fname): os.unlink(fname)



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import egglib, unittest, pathlib
path = pathlib.Path(__file__).parent / '..' / 'data'

class Clustal_test(unittest.TestCase):
    def test_clustal_T(self):
        cnt = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)
        aln = egglib.wrappers.clustal(cnt, verbose=False, threads=8, keep_order=True)
        Lts_cnt=cnt.find('Lotus')
        Lts_aln=aln.find('Lotus')
        seq_cnt=Lts_cnt.sequence.string()
        seq_aln=Lts_aln.sequence.string()
        seq_=seq_aln.replace("-","")
        n_miss=seq_aln.count('-')
            
        self.assertIsInstance(aln, egglib.Align)
        self.assertNotEqual(len(seq_cnt), len(seq_aln))
        self.assertEqual(len(seq_cnt), (len(seq_aln)-n_miss))
        self.assertEqual(seq_, seq_cnt)

    def test_clustal_E(self):
        cnt = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)
        cache = egglib.wrappers.paths['clustal']
        egglib.wrappers.paths['clustal'] = None
        with self.assertRaises(RuntimeError):
            aln = egglib.wrappers.clustal(cnt, verbose=False, threads=8, keep_order=True)
        egglib.wrappers.paths['clustal'] = cache

        cnt_ref = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, cls=None)
        with self.assertRaises(TypeError):
            aln = egglib.wrappers.clustal(cnt, ref=cnt_ref, verbose=False, threads=8, keep_order=True)

        aln_e = egglib.io.from_fasta(str(path / 'error.fas'), egglib.alphabets.DNA, cls=egglib.Align) #empty file
        with self.assertRaises(ValueError):
            aln = egglib.wrappers.clustal(cnt_ref, ref=aln_e, verbose=False, threads=8, keep_order=True)
        
        cnt=egglib.io.from_fasta(str(path / 'cds_e.fas'), egglib.alphabets.DNA, labels=True)
        with self.assertRaises(ValueError):
            aln = egglib.wrappers.clustal(cnt, verbose=False, threads=8, keep_order=True)
        
        aln_e = egglib.io.from_fasta(str(path / 'error.fas'), egglib.alphabets.DNA, cls=egglib.Align) #empty file
        with self.assertRaises(ValueError):
            aln = egglib.wrappers.clustal(aln_e, verbose=False, threads=8, keep_order=True)

        aln_ref = egglib.io.from_fasta(str(path / 'cds_clust.fas'), egglib.alphabets.DNA, cls=egglib.Align) #empty file
        with self.assertRaises(ValueError):
            aln = egglib.wrappers.clustal(aln_e, ref=aln_ref, verbose=False, threads=8, keep_order=True)

        Lts_cnt=['e', 'r', 'r', 'o', 'r']
        with self.assertRaises(AttributeError):
            aln = egglib.wrappers.clustal(Lts_cnt,verbose=False, threads=8, keep_order=True)

        cnt = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)

        with self.assertRaises(ValueError):
                aln= egglib.wrappers.clustal(cnt,verbose=False, threads=8, use_kimura=True, keep_order=True)
        
        with self.assertRaises(ValueError):
                aln= egglib.wrappers.clustal(cnt ,verbose=False, num_iter=-10, threads=8)

        with self.assertRaises(ValueError):
                aln= egglib.wrappers.clustal(cnt ,verbose=False, num_iter='100', threads=8)

        with self.assertRaises(ValueError):
                aln= egglib.wrappers.clustal(cnt ,verbose=False, threads=-8)


        aln_ref = egglib.io.from_fasta(str(path / 'codon_align.fas'), egglib.alphabets.DNA)
        with self.assertRaises(RuntimeError):
            aln = egglib.wrappers.clustal(cnt, ref=aln_ref, verbose=False, threads=8, keep_order=True)

    def test_clustal_prot(self):
        c1 = egglib.io.from_fasta(str(path / 'cds.fas'), egglib.alphabets.DNA, labels=True)
        c2 = egglib.io.from_fasta(str(path / 'prot_data.fas'), alphabet=egglib.alphabets.protein)
        a1 = egglib.wrappers.clustal(c1)
        a2 = egglib.wrappers.clustal(c2)
